# -*- coding: utf-8 -*-
"""abydos

Abydos NLP/IR library by Christopher C. Little

This library contains code I'm using for research, in particular dissertation
research & experimentation.

Further documentation to come...

Copyright 2014 by Christopher C. Little.

Abydos is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Abydos is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Abydos. If not, see <http://www.gnu.org/licenses/>.
"""
